package com.gic.spark.tag;

import com.alibaba.fastjson.JSONObject;
import com.alibaba.fastjson.TypeReference;
import com.gic.spark.application.SparkEnvManager;
import com.gic.spark.datasource.DataSourceManager;
import com.gic.spark.datasource.entity.DataSourceEntity;
import com.gic.spark.datasource.entity.DataSourceSharding;
import com.gic.spark.datasource.mysql.MysqlDatasource;
import com.gic.spark.datasource.mysql.MysqlRddManager;
import com.gic.spark.entity.SceneCrowdDTO;
import com.gic.spark.entity.TagConditionDTO;
import com.gic.spark.entity.TagConditionGroupDTO;
import com.gic.spark.entity.TagGroupInfo;
import com.gic.spark.entity.request.AbstractFilterRequest;
import com.gic.spark.entity.table.TabSceneCrowd;
import com.gic.spark.filter.*;
import com.gic.spark.util.AppEnvUtil;
import com.gic.spark.util.DingtalkMessageUtil;
import com.gic.spark.util.EsRequestUtil;
import com.google.common.base.Joiner;
import com.google.common.collect.Lists;
import org.apache.commons.lang.StringUtils;
import org.apache.spark.TaskContext;
import org.apache.spark.api.java.*;
import org.apache.spark.sql.Column;
import org.apache.spark.sql.Dataset;
import org.apache.spark.sql.Row;
import org.apache.spark.sql.SparkSession;
import org.elasticsearch.hadoop.cfg.ConfigurationOptions;
import org.elasticsearch.hadoop.cfg.PropertiesSettings;
import org.elasticsearch.hadoop.cfg.Settings;
import org.elasticsearch.spark.ESShardPartitioner;
import org.elasticsearch.spark.EsRddFactory;
import org.elasticsearch.spark.cfg.SparkSettings;
import org.elasticsearch.spark.cfg.SparkSettingsManager;
import org.elasticsearch.spark.rdd.EsRDDWriter;
import scala.Tuple2;
import scala.collection.JavaConversions;
import java.util.*;
import org.apache.spark.api.java.Optional;

import static com.gic.spark.datasource.mysql.MysqlDatasource.JDBC_OPTIONS;

/**
 * @description:
 * @author: wangxk
 * @date: 2020/4/15
 */
public class TagProcess {

    private List<SceneCrowdDTO>sceneCrowdDTOList=new ArrayList();
    private MysqlRddManager mysqlRddManager;
    private DataSourceSharding memberDatasource;
    private boolean isProduction;
    private static TagProcess instance;
    public static TagProcess getInstance(){
        if(null==instance){
            instance=new TagProcess();
        }
        return instance;
    }
    private TagProcess(){}

    public boolean isProduction() {
        return isProduction;
    }

    public void init(boolean isProd){
        this.isProduction=isProd;
        MysqlDatasource member4Datasource = new MysqlDatasource();

        if(isProduction){
            member4Datasource.setJdbcUrl("jdbc:mysql://10.0.0.6:3306/gic-member4.0" + JDBC_OPTIONS);
            member4Datasource.setUser("cdb_outerroot");
            member4Datasource.setPassword("@09ui%sbc09");
            memberDatasource = new DataSourceSharding("member_sharding_4.0_prod", "tab_member_user");
        }else{
            member4Datasource.setJdbcUrl("jdbc:mysql://10.0.0.6:3306/gic-member4.0" + JDBC_OPTIONS);
            member4Datasource.setUser("cdb_outerroot");
            member4Datasource.setPassword("@09ui%sbc09");
            memberDatasource = new DataSourceSharding("member_sharding_4.0_test", "tab_member_user");
        }
        mysqlRddManager=member4Datasource.buildRddManager();

        Dataset<TabSceneCrowd>sceneCrowdDataset=mysqlRddManager.getPojo("tab_scene_crowd", TabSceneCrowd.class,null)
                                                                .filter(new Column("valid_flag").equalTo(1));
        List<TabSceneCrowd> sceneCrowdList=sceneCrowdDataset.collectAsList();
        for(TabSceneCrowd sceneCrowd:sceneCrowdList){
            LinkedList<TagConditionGroupDTO>conditionGroupDTOList=JSONObject.parseObject(sceneCrowd.getTag_Condition_Group_Info(),  new TypeReference<LinkedList<TagConditionGroupDTO>>(){});
            sceneCrowdDTOList.add(new SceneCrowdDTO(sceneCrowd,conditionGroupDTOList));
        }

    }

    public void setEnterpriseId(List<Integer>enterpriseIdList){
        for(SceneCrowdDTO dto:sceneCrowdDTOList){
            if(!enterpriseIdList.contains(dto.getEnterprise_Id())){
                sceneCrowdDTOList.remove(dto);
            }
        }
    }
    public void setTagGroupId(List<Integer>tagGroupIdList){
        for(SceneCrowdDTO dto:sceneCrowdDTOList){
            if(!tagGroupIdList.contains(dto.getId())){
                sceneCrowdDTOList.remove(dto);
            }
        }
    }

    public void process(boolean extractData){

        Map<Integer, List<TagProcessEntity>> tagGroupByEnterpriseMap = new HashMap<>();
        Map<Long, TagFilter> tagIdToFilterMap = new HashMap();
        for(SceneCrowdDTO sceneCrowdDTO:sceneCrowdDTOList){
            if(sceneCrowdDTO.getReal_Time()==1){
                LinkedList<TagConditionGroupDTO> conditionGroupDTOS=sceneCrowdDTO.getConditionGroupDTOList();
                for(int i=0;i<conditionGroupDTOS.size();i++){

                    TagProcessEntity entity=new TagProcessEntity();
                    entity.enterpriseId=sceneCrowdDTO.getEnterprise_Id();
                    entity.tagGroupId=sceneCrowdDTO.getId();
                    entity.level=i+1;

                    entity.tagList=conditionGroupDTOS.get(i).getConditionInfos();
                    for(TagConditionDTO conditionDTO:entity.tagList){//将tag同filter进行映射
                        TagFilter tagFilter=TagFilterFactory.getInstance().getTagFilter(conditionDTO.getTagEsFieldName());
                        if(null!=tagFilter){
                            tagIdToFilterMap.put(conditionDTO.getTagId(),tagFilter);
                        }
                    }

                    if(!tagGroupByEnterpriseMap.containsKey(sceneCrowdDTO.getEnterprise_Id())){
                        tagGroupByEnterpriseMap.put(sceneCrowdDTO.getEnterprise_Id(),new ArrayList());
                    }
                    tagGroupByEnterpriseMap.get(sceneCrowdDTO.getEnterprise_Id()).add(entity);
                }
            }
        }

        //准备标签需要的数据
        for (Map.Entry<Integer, List<TagProcessEntity>> enterpriseTagEntry : tagGroupByEnterpriseMap.entrySet()) {
            for (TagProcessEntity processEntity : enterpriseTagEntry.getValue()) {
                for (TagConditionDTO conditionDTO : processEntity.tagList) {
                    if (tagIdToFilterMap.containsKey(conditionDTO.getTagId())) {
                        for (DataSourceEntity sourceEntity : tagIdToFilterMap.get(conditionDTO.getTagId()).necessarySourceList()) {
                            DataSourceManager.getInstance().addSourceEntity(sourceEntity, enterpriseTagEntry.getKey().intValue());
                        }
                    }
                }
            }
            DataSourceManager.getInstance().addSourceEntity(memberDatasource,enterpriseTagEntry.getKey().intValue());
        }

        if (extractData) {
            DataSourceManager.getInstance().init(isProduction);
            DataSourceManager.getInstance().extractDataToDatabase();
            DingtalkMessageUtil.sendAlertMessage("extractData to hive finish !");
        }

        //处理标签数据
        for (Map.Entry<Integer, List<TagProcessEntity>> enterpriseTagEntry : tagGroupByEnterpriseMap.entrySet()) {
            Integer enterpriseId=enterpriseTagEntry.getKey();
            List<JavaPairRDD<Integer, String>> filterRddList = new ArrayList();
            for(TagProcessEntity entity:enterpriseTagEntry.getValue()){
                for(TagConditionDTO conditionDTO:entity.tagList){
                    if(tagIdToFilterMap.containsKey(conditionDTO.getTagId())){
                        TagFilter tagFilter=tagIdToFilterMap.get(conditionDTO.getTagId());
                        AbstractFilterRequest filterRequest=TagValueParser.parseFilterValue(conditionDTO);

                        final String groupId = entity.tagGroupId + "_" + conditionDTO.getTagId() + "_" + entity.level;

                        JavaPairRDD<Integer, String> rdd = tagFilter.filterValidMember(enterpriseId,filterRequest).mapToPair(data-> Tuple2.apply(data,groupId));
                        filterRddList.add(rdd);
                    }
                }
            }

            if(filterRddList.size()>0){
                JavaPairRDD<Integer,String>memberGroupRdd=null;
                for(JavaPairRDD<Integer,String>rdd:filterRddList){
                    if(null==memberGroupRdd){
                        rdd=memberGroupRdd;
                    }else{
                        memberGroupRdd.union(rdd);
                    }
                }
                JavaPairRDD<Integer,Integer>memberUserRdd=memberDatasource.getDatasetByEnterpriseId(enterpriseId).select("id").javaRDD()
                                            .mapToPair(data->Tuple2.apply((Integer)data.getAs("id"),(Integer)data.getAs("id")))
                                            .reduceByKey((x,y)->x);

                JavaPairRDD<Integer,String>updateMemberGroupRdd=memberUserRdd.leftOuterJoin(memberGroupRdd.groupByKey())
                                                                            .mapPartitionsToPair(data->{
                                                                                List<Tuple2<Integer,String>> result=new ArrayList();
                                                                                while (data.hasNext()){
                                                                                    Tuple2<Integer, Optional<Iterable<String>>> tp2=data.next()._2();
                                                                                    Integer mcuId=tp2._1();
                                                                                    String groupId=tp2._2().isPresent()? Joiner.on(" ").join(Lists.newArrayList(tp2._2().get())):"";
                                                                                    JSONObject json = new JSONObject();
                                                                                    json.put("mcuId",mcuId);
                                                                                    json.put("tagGroup",groupId);
                                                                                    result.add(Tuple2.apply(mcuId,json.toString()));
                                                                                }
                                                                                return result.iterator();
                                                                            });

                JavaPairRDD<Integer,String>cacheMemberGroupRdd=updateMemberGroupRdd.cache();

                updateIndex(cacheMemberGroupRdd,enterpriseId);

                saveToHive(cacheMemberGroupRdd,enterpriseId);

                cacheMemberGroupRdd.unpersist();
            }
        }


    }

    private void saveToHive(JavaPairRDD<Integer, String> updateMemberGroup,Integer enterpriseId){
        JavaRDD<TagGroupInfo> tagGroupInfoRDD=updateMemberGroup.mapPartitions(data->{
                                                List<TagGroupInfo>tagGroupInfoList=new ArrayList();
                                                while(data.hasNext()){
                                                    Tuple2<Integer,String>tp2=data.next();
                                                    JSONObject json = JSONObject.parseObject(tp2._2());
                                                    if(StringUtils.isNotEmpty(json.getString("tagGroup"))){
                                                        tagGroupInfoList.add(new TagGroupInfo(json.getInteger("mcuId"),enterpriseId,json.getString("tagGroup")));
                                                    }
                                                }
                                                return tagGroupInfoList.iterator();
                                            });
        SparkSession sparkSession=SparkEnvManager.getInstance().getSparkSession();
        Dataset<Row>tagGroupInfoDataset=sparkSession.createDataFrame(tagGroupInfoRDD,TagGroupInfo.class);
        tagGroupInfoDataset.createOrReplaceTempView("tag_group_info_tmp");
        sparkSession.sql(String.format("INSERT overwrite table tag.tag_group_info PARTITION(enterprise_id = '%s') SELECT mcu_id,enterprise_id,tag_group_info from tag_info_tmp", enterpriseId));

    }

    private void updateIndex(JavaPairRDD<Integer, String> updateMemberGroup, Integer enterpriseId) {

        Map<String, String> conf = new HashMap();
        String indexName = EsRequestUtil.getESIindexName(enterpriseId);
        System.out.println("indexName===>" + indexName);
        conf.put("es.resource", indexName + "/mapper_type");
        conf.put("es.nodes", AppEnvUtil.ES_NODES);
        conf.put("es.mapping.date.rich", "false");
        conf.put("es.index.auto.create", "true");
        conf.put("es.mapping.id", "memberId");
        conf.put("es.write.operation", "upsert");
        conf.put("es.batch.size.bytes", "1024k");
        conf.put("es.batch.size.entries", "1500");
        conf.put(ConfigurationOptions.ES_INPUT_JSON, "true");

        SparkSettings sparkCfg = new SparkSettingsManager().load(SparkEnvManager.getInstance().getSparkContext().getConf());
        Settings cfg = new PropertiesSettings().load(sparkCfg.save());
        cfg.merge(conf);
        String settings = cfg.save();

        ESShardPartitioner partitioner = new ESShardPartitioner(settings);

        updateMemberGroup.partitionBy(partitioner).map(data ->data._2()).foreachPartition(data -> {
            Settings newSettings = new PropertiesSettings().load(settings);
            EsRDDWriter<String> writer = EsRddFactory.createJsonWriter(newSettings.save(), false);
            writer.write(TaskContext.get(), JavaConversions.asScalaIterator(data));
        });
    }

}
